#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import argparse
import re
import sys
import genhub


class FeatureFormatter(object):
    """Load features from GFF3, parse and (re-)attach accession numbers."""

    def __init__(self, instream, source):
        self.instream = instream
        self.source = source

        self.id2type = dict()
        self.id2acc = dict()
        self.filters = ['\tregion\t', '\tmatch\t', '\tcDNA_match\t',
                        '##species']

    def __iter__(self):
        for line in self.instream:
            line = line.rstrip()
            if self.match_filter(line):
                continue
            if self.pseudogenic_cds(line):
                continue

            self.parse_type(line)
            line = self.parse_gene(line)
            line = self.parse_transcript(line)
            line = self.parse_vdj(line)
            line = self.parse_feature(line)

            yield line

    def parse_type(self, line):
        fields = line.split('\t')
        if len(fields) != 9:
            return False

        ftype = fields[2]
        attributes = fields[8]
        idmatch = re.search('ID=([^;\n]+)', attributes)
        if idmatch:
            featureid = idmatch.group(1)
            self.id2type[featureid] = ftype

    def match_filter(self, line):
        for filt in self.filters:
            if filt in line:
                return True
        return False

    def pseudogenic_cds(self, line):
        """
        Test whether the given entry is a pseudogene-associated CDS.

        We want to ignore these!
        """
        if self.source == 'tair':
            return False

        fields = line.split('\t')
        if len(fields) != 9:
            return False

        ftype = fields[2]
        attributes = fields[8]
        if ftype != 'CDS':
            return False

        parentmatch = re.search('Parent=([^;\n]+)', attributes)
        assert parentmatch, attributes
        parentid = parentmatch.group(1)
        if self.id2type[parentid] == 'pseudogene':
            return True
        return False

    def parse_gene(self, line):
        """Parse accession for gene features."""
        fields = line.split('\t')
        if len(fields) != 9:
            return line

        ftype = fields[2]
        attributes = fields[8]
        if ftype != 'gene' or '\tAEGeAn::tidygff3\t' in line:
            return line

        accmatch = None
        if self.source in ['refseq', 'ncbi_flybase']:
            accmatch = re.search('GeneID:([^;,\n]+)', line)
        elif self.source == 'crg':
            accmatch = re.search('ID=([^;\n]+)', line)
        elif self.source in ['pdom', 'tair', 'beebase']:
            accmatch = re.search('Name=([^;\n]+)', line)
        else:
            pass
        assert accmatch, 'unable to parse gene accession: %s' % line
        accession = accmatch.group(1)

        idmatch = re.search('ID=([^;\n]+)', attributes)
        if idmatch:
            geneid = idmatch.group(1)
            self.id2acc[geneid] = accession
        else:
            print('Warning: gene has no ID: %s' % attributes, file=sys.stderr)

        return line + ';accession=' + accession

    def parse_transcript(self, line):
        """Parse accession for transcript features."""
        fields = line.split('\t')
        if len(fields) != 9:
            return line

        ftype = fields[2]
        attributes = fields[8]
        if ftype not in ['mRNA', 'tRNA', 'rRNA', 'ncRNA', 'miRNA', 'snRNA',
                         'snoRNA', 'transcript', 'primary_transcript']:
            return line

        accmatch = None
        idmatch = None
        if self.source in ['refseq', 'ncbi_flybase']:
            accmatch = re.search('transcript_id=([^;\n]+)', attributes)
            idmatch = re.search('GeneID:([^;,\n]+)', attributes)
        elif self.source in ['crg', 'pdom']:
            accmatch = re.search('ID=([^;\n]+)', attributes)
        elif self.source in ['beebase', 'tair', 'am10']:
            accmatch = re.search('Name=([^;\n]+)', attributes)
        else:
            pass
        assert accmatch or idmatch, \
            'unable to parse transcript accession: %s' % line
        if accmatch:
            accession = accmatch.group(1)
        else:
            accession = '%s:%s' % (idmatch.group(1), ftype)

        rnaidmatch = re.search('ID=([^;\n]+)', attributes)
        if rnaidmatch:
            rnaid = rnaidmatch.group(1)
            self.id2acc[rnaid] = accession
        else:
            print('Warning: RNA has no ID: %s' % attributes, file=sys.stderr)

        return line + ';accession=' + accession

    def parse_vdj(self, line):
        """Parse accessions for features of V(D)J genes."""
        fields = line.split('\t')
        if len(fields) != 9:
            return line

        ftype = fields[2]
        if ftype not in ['V_gene_segment', 'D_gene_segment', 'J_gene_segment',
                         'C_gene_segment']:
            return line

        vdjid = re.search('ID=([^;\n]+)', line).group(1)
        accession = re.search('GeneID:([^;,\n]+)', line).group(1)
        self.id2acc[vdjid] = accession

        return line + ';accession=' + accession

    def parse_feature(self, line):
        """Parse accession for exons, introns, and coding sequences"""
        fields = line.split('\t')
        if len(fields) != 9:
            return line

        ftype = fields[2]
        if ftype not in ['exon', 'intron', 'CDS']:
            return line

        parentid = re.search('Parent=([^;\n]+)', line).group(1)
        if self.source == 'tair':
            for pid in parentid.split(','):
                if 'RNA' in pid:
                    parentid = pid
        assert ',' not in parentid, parentid
        assert parentid in self.id2acc, parentid
        accession = self.id2acc[parentid]
        return line + ';accession=' + accession


def parse_args():
    """Define the command-line interface."""
    sources = ['refseq', 'ncbi_flybase', 'beebase', 'crg', 'pdom', 'tair',
               'am10']
    desc = 'Filter features and parse accession values'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-v', '--version', action='version',
                        version='GenHub v%s' % genhub.__version__)
    parser.add_argument('-o', '--outfile', type=argparse.FileType('w'),
                        default=sys.stdout)
    parser.add_argument('-p', '--prefix', default=None, metavar='PFX',
                        help='attach the given prefix to each sequence ID')
    parser.add_argument('--source', default='refseq', choices=sources,
                        help='data source; default is "refseq"')
    parser.add_argument('gff3', type=argparse.FileType('r'))
    return parser.parse_args()


def format_prefix(line, prefix):
    """Apply a prefix to each sequence ID."""
    if len(line.split('\t')) == 9:
        return prefix + line
    elif line.startswith('##sequence-region'):
        return re.sub('##sequence-region(\s+)(\S+)',
                      '##sequence-region\g<1>%s\g<2>' % args.prefix, line)


def main():
    args = parse_args()
    formatter = FeatureFormatter(args.gff3, args.source)
    for line in formatter:
        if args.prefix:
            line = format_prefix(line, args.prefix)
        print(line, file=args.outfile)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import filecmp
import re
import subprocess
import sys
import genhub


def mrna_exons(instream, convert=False, keepMrnas=False, usecds=False):
    """
    Parse exons from a data stream in GFF3 format.

    - If `convert` is true, convert the feature type of the exons to `mRNA`,
      creating a mature mRNA multi-feature.
    - If `keepMrnas` is true, return mRNA features in addition to exon
      features.
    - If `usecds` is true, parse exon structure from `CDS` features instead of
      `exon` features.
    """
    mrnaids = {}
    for line in instream:
        line = line.rstrip()
        fields = line.split('\t')
        if len(fields) != 9:
            continue
        exontype = 'exon'
        if usecds:
            exontype = 'CDS'

        if fields[2] == 'mRNA':
            mrnaid = re.search('ID=([^;\n]+)', fields[8]).group(1)
            accmatch = re.search('accession=([^;\n]+)', fields[8])
            assert accmatch, 'Unable to parse mRNA accession: %s' % fields[8]
            mrnaacc = accmatch.group(1)
            mrnaids[mrnaid] = 1
            if not convert and keepMrnas:  # pragma: no cover
                fields[8] = re.sub('Parent=[^;\n]+;*', '', fields[8])
                yield '\t'.join(fields)

        elif fields[2] == exontype:
            parentid = re.search('Parent=([^;\n]+)', fields[8]).group(1)
            fields[7] = '.'
            if parentid in mrnaids:
                if convert:
                    fields[2] = 'mRNA'
                    fields[8] = re.sub('ID=[^;\n]+;*', '', fields[8])
                    fields[8] = fields[8].replace('Parent=', 'ID=')
                    if 'accession=' not in fields[8]:  # pragma: no cover
                        fields[8] += ';accession=' + mrnaacc
                else:
                    if not keepMrnas:  # pragma: no cover
                        fields[8] = re.sub('Parent=[^;\n]+;*', '', fields[8])
                yield '\t'.join(fields)


def mature_mrna_intervals(db, logstream=sys.stderr):
    """
    Parse gene model structures and create mRNA mutli-features.

    Extracting the sequence of a pre-mRNA is trivial, but extracting the
    sequence of a mature mRNA (sans introns) requires some additional work.
    This function creates a new GFF3 file containing mRNA multi-features,
    enabling sequence extraction via xtractore.
    """
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'calculating mature mRNA intervals'
        print(logmsg, file=logstream)
    specdir = '%s/%s' % (db.workdir, db.label)

    infile = '%s/%s.gff3' % (specdir, db.label)
    outfile = '%s/%s.mrnas.temp' % (specdir, db.label)
    usecds = False
    if repr(db) in ['BeeBase', 'OGS1.0']:
        usecds = True
    with open(infile, 'r') as instream, open(outfile, 'w') as outstream:
        for exon in mrna_exons(instream, convert=True, usecds=usecds):
            print(exon, file=outstream)

    infile = '%s/%s.ilocus.mrnas.gff3' % (specdir, db.label)
    outfile = '%s/%s.ilocus.mrnas.temp' % (specdir, db.label)
    with open(infile, 'r') as instream, open(outfile, 'w') as outstream:
        for exon in mrna_exons(instream, convert=True, usecds=usecds):
            print(exon, file=outstream)

    inpatterns = ['%s/%s.mrnas.temp', '%s/%s.ilocus.mrnas.temp']
    outpatterns = ['%s/%s.all.mrnas.gff3', '%s/%s.mrnas.gff3']
    for inpattern, outpattern in zip(inpatterns, outpatterns):
        inf = inpattern % (specdir, db.label)
        otf = outpattern % (specdir, db.label)
        command = 'gt gff3 -retainids -sort -tidy -force -o %s %s' % (otf, inf)
        cmd = command.split(' ')
        proc = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                                universal_newlines=True)
        _, stderr = proc.communicate()
        for line in stderr.split('\n'):  # pragma: no cover
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               line != '':
                print(line, file=logstream)


def sequences(db, logstream=sys.stderr):
    if logstream is not None:  # pragma: no cover
        logmsg = '[GenHub: %s] ' % db.config['species']
        logmsg += 'extracting pre-mRNA and mature mRNA sequences'
        print(logmsg, file=logstream)
    specdir = '%s/%s' % (db.workdir, db.label)

    # All pre-mRNA sequences
    gff3infile = '%s/%s.gff3' % (specdir, db.label)
    fastainfile = '%s/%s.gdna.fa' % (specdir, db.label)
    outfile = '%s/%s.all.pre-mrnas.fa' % (specdir, db.label)
    command = 'xtractore --type=mRNA --outfile=%s ' % outfile
    command += '%s %s' % (gff3infile, fastainfile)
    cmd = command.split(' ')
    subprocess.check_call(cmd)

    # All mature mRNA sequences
    gff3infile = '%s/%s.all.mrnas.gff3' % (specdir, db.label)
    fastainfile = '%s/%s.gdna.fa' % (specdir, db.label)
    outfile = '%s/%s.all.mrnas.fa' % (specdir, db.label)
    command = 'xtractore --type=mRNA --outfile=%s ' % outfile
    command += '%s %s' % (gff3infile, fastainfile)
    cmd = command.split(' ')
    subprocess.check_call(cmd)

    # Representative pre-mRNA sequences
    idfile = '%s/%s.mrnas.txt' % (specdir, db.label)
    seqfile = '%s/%s.all.pre-mrnas.fa' % (specdir, db.label)
    outfile = '%s/%s.pre-mrnas.fa' % (specdir, db.label)
    with open(idfile, 'r') as idstream, \
            open(seqfile, 'r') as seqstream, \
            open(outfile, 'w') as outstream:
        for defline, seq in genhub.fasta.select(idstream, seqstream):
            print(defline, file=outstream)
            genhub.fasta.format(seq, outstream=outstream)

    # Representative mature mRNA sequences
    idfile = '%s/%s.mrnas.txt' % (specdir, db.label)
    seqfile = '%s/%s.all.mrnas.fa' % (specdir, db.label)
    outfile = '%s/%s.mrnas.fa' % (specdir, db.label)
    with open(idfile, 'r') as idstream, \
            open(seqfile, 'r') as seqstream, \
            open(outfile, 'w') as outstream:
        for defline, seq in genhub.fasta.select(idstream, seqstream):
            print(defline, file=outstream)
            genhub.fasta.format(seq, outstream=outstream)


# -----------------------------------------------------------------------------
# Driver function
# -----------------------------------------------------------------------------

def prepare(db, logstream=sys.stderr):  # pragma: no cover
    mature_mrna_intervals(db, logstream=logstream)
    sequences(db, logstream=logstream)


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_mature_mrna_intervals():
    """Define mature mRNA intervals"""
    config = genhub.test_registry.genome('Atha')
    d = genhub.refseq.RefSeqDB('Atha', config, workdir='testdata/demo-workdir')
    mature_mrna_intervals(d, logstream=None)

    outfile = 'testdata/demo-workdir/Atha/Atha.all.mrnas.gff3'
    testfile = 'testdata/gff3/atha-all-mrnas.gff3'
    assert filecmp.cmp(outfile, testfile), 'mature mRNA interval ID failed'

    outfile = 'testdata/demo-workdir/Atha/Atha.mrnas.gff3'
    testfile = 'testdata/gff3/atha-mrnas.gff3'
    assert filecmp.cmp(outfile, testfile), 'mature mRNA interval ID failed'

    config = genhub.test_registry.genome('Dnov')
    d = genhub.beebase.BeeBaseDB('Dnov', config,
                                 workdir='testdata/demo-workdir')
    mature_mrna_intervals(d, logstream=None)

    outfile = 'testdata/demo-workdir/Dnov/Dnov.all.mrnas.gff3'
    testfile = 'testdata/gff3/dnov-all-mrnas.gff3'
    assert filecmp.cmp(outfile, testfile), 'mature mRNA interval ID failed'

    outfile = 'testdata/demo-workdir/Dnov/Dnov.mrnas.gff3'
    testfile = 'testdata/gff3/dnov-mrnas.gff3'
    assert filecmp.cmp(outfile, testfile), 'mature mRNA interval ID failed'


def test_mrna_sequences():
    """Extract mRNA sequences"""
    config = genhub.test_registry.genome('Atha')
    d = genhub.refseq.RefSeqDB('Atha', config, workdir='testdata/demo-workdir')
    sequences(d, logstream=None)

    outfile = 'testdata/demo-workdir/Atha/Atha.all.pre-mrnas.fa'
    testfile = 'testdata/fasta/atha-all-pre-mrnas.fa'
    with open(outfile, 'r') as out, open(testfile, 'r') as test:
        assert genhub.fasta.compare(out, test) is True, \
            'all pre-mRNA seq extraction failed'

    outfile = 'testdata/demo-workdir/Atha/Atha.pre-mrnas.fa'
    testfile = 'testdata/fasta/atha-pre-mrnas.fa'
    with open(outfile, 'r') as out, open(testfile, 'r') as test:
        assert genhub.fasta.compare(out, test) is True, \
            'pre-mRNA seq extraction failed'

    outfile = 'testdata/demo-workdir/Atha/Atha.all.pre-mrnas.fa'
    testfile = 'testdata/fasta/atha-all-pre-mrnas.fa'
    with open(outfile, 'r') as out, open(testfile, 'r') as test:
        assert genhub.fasta.compare(out, test) is True, \
            'all mRNA seq extraction failed'

    outfile = 'testdata/demo-workdir/Atha/Atha.mrnas.fa'
    testfile = 'testdata/fasta/atha-mrnas.fa'
    with open(outfile, 'r') as out, open(testfile, 'r') as test:
        assert genhub.fasta.compare(out, test) is True, \
            'mature mRNA seq extraction failed'



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------


class HiLocus(object):

    def __init__(self, record):
        self._rawdata = record
        values = record.split('\t')
        self.iloci = values[2].split(',')
        self.species = values[3].split(',')
        self.iloci_by_species = dict()
        for locus in self.iloci:
            species = locus[6:10]
            if species not in self.iloci_by_species:
                self.iloci_by_species[species] = list()
            self.iloci_by_species[species].append(locus)
        self.simple_iloci = None

    def in_clade(self, clade):
        assert self.simple_iloci
        reps = list()
        for species in clade:
            copynumber = len(self.iloci_by_species[species])
            assert copynumber > 0
            if copynumber > 1:
                continue
            ilocus = self.iloci_by_species[species]
            if ilocus not in self.simple_iloci:
                continue
            reps.append(ilocus)
        return reps



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------
"""
Process a GFF3 file, and for any feature that has an ID but lacks a `Name`
attribute, copy the ID attribute to the Name attribute.
"""

from __future__ import print_function
import re
import sys

for line in sys.stdin:
    line = line.rstrip()
    idmatch = re.search("ID=([^;\n]+)", line)
    namematch = re.search("Name=([^;\n]+)", line)
    if idmatch and not namematch:
        line += ";Name=%s" % idmatch.group(1)
    print(line)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""
Retrieve and format data from The Arabidopsis Information Resource.

GenomeDB implementation for data residing in TAIR. Currently tested only on
TAIR6, but presumably trivial to extend to other versions using Att6.yml as a
template.
"""

from __future__ import print_function
import filecmp
import gzip
import os
import re
import subprocess
import sys
import genhub


class TairDB(genhub.genomedb.GenomeDB):

    def __init__(self, label, conf, workdir='.'):
        super(TairDB, self).__init__(label, conf, workdir)

        assert self.config['source'] == 'tair'
        assert 'version' in self.config
        assert str(self.config['version']) == '6'
        assert 'species' in self.config
        assert self.config['species'] == 'Arabidopsis thaliana'

        self.specbase = 'ftp://ftp.arabidopsis.org/home/tair'
        self.format_gdna = self.format_fasta
        self.format_prot = self.format_fasta

    def __repr__(self):
        return 'TAIR' + self.version

    @property
    def version(self):
        return str(self.config['version'])

    @property
    def gdnafilename(self):
        return self.config['gdna_filename']

    @property
    def gff3filename(self):
        return self.config['gff3_filename']

    @property
    def protfilename(self):
        return self.config['prot_filename']

    @property
    def gdnaurl(self):
        return self.config['gdna_url']

    @property
    def gff3url(self):
        return self.config['gff3_url']

    @property
    def proturl(self):
        return self.config['prot_url']

    def format_fasta(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            # No processing required.
            # If any is ever needed again, do it here.
            print(line, end='', file=outstream)

    def format_gff3(self, logstream=sys.stderr, debug=False):
        cmds = list()
        cmds.append('gunzip -c %s' % self.gff3path)
        if 'annotfilter' in self.config:  # pragma: no cover
            excludefile = self.filter_file()
            cmds.append('grep -vf %s' % excludefile.name)
        cmds.append('sed "s/Index=/index=/"')
        cmds.append('tidygff3')
        cmds.append('genhub-format-gff3.py --source tair -')
        cmds.append('gt gff3 -sort -tidy -o %s -force' % self.gff3file)

        commands = ' | '.join(cmds)
        if debug:  # pragma: no cover
            print('DEBUG: running command: %s' % commands, file=logstream)
        proc = subprocess.Popen(commands, shell=True, universal_newlines=True,
                                stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        for line in stderr.split('\n'):  # pragma: no cover
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               'more than one pseudogene attribute' not in line and \
               line != '':
                print(line, file=logstream)
        assert proc.returncode == 0, \
            'annot cleanup command failed: %s' % commands
        if 'annotfilter' in self.config:  # pragma: no cover
            os.unlink(excludefile.name)

    def gff3_protids(self, instream):
        protids = dict()
        for line in instream:
            if '\tmRNA\t' not in line:
                continue
            idmatch = re.search('Name=([^;\n]+)', line)
            assert idmatch, 'cannot parse protein ID: ' + line
            protid = idmatch.group(1)
            assert protid not in protids, protid
            protids[protid] = True
            yield protid

    def protein_mapping(self, instream):
        locusid2name = dict()
        gene2loci = dict()
        for line in instream:
            fields = line.split('\t')
            if len(fields) != 9:
                continue
            feattype = fields[2]
            attrs = fields[8]

            if feattype == 'locus':
                idmatch = re.search('ID=([^;\n]+);.*Name=([^;\n]+)', attrs)
                if idmatch:
                    locusid = idmatch.group(1)
                    locusname = idmatch.group(2)
                    locusid2name[locusid] = locusname
            elif feattype == 'gene':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                if idmatch:
                    geneid = idmatch.group(1)
                    ilocusid = idmatch.group(2)
                    gene2loci[geneid] = ilocusid
                else:  # pragma: no cover
                    print('Unable to parse gene and iLocus IDs: %s' % attrs,
                          file=sys.stderr)
            elif feattype == 'mRNA':
                idmatch = re.search('Parent=([^;\n]+);Name=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse mRNA and gene IDs: %s' % attrs.rstrip()
                geneid = idmatch.group(1)
                mrnaid = idmatch.group(2)
                ilocusid = gene2loci[geneid]
                ilocusname = locusid2name[ilocusid]
                yield mrnaid, ilocusname


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_gdna_format():
    """TAIR6 gDNA formatting"""
    conf = genhub.test_registry.genome('Att6')
    tair_db = TairDB('Att6', conf, workdir='testdata/demo-workdir')
    tair_db.preprocess_gdna(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Att6/Att6.gdna.fa'
    testoutfile = 'testdata/fasta/tair6-gdna-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'TAIR6 gDNA formatting failed'

    assert repr(tair_db) == 'TAIR6'
    ftpbase = ('ftp://ftp.arabidopsis.org/home/tair/Sequences/'
               'whole_chromosomes/OLD/chr%d.fas')
    urls = [ftpbase % x for x in (1, 2, 3, 4, 5)]
    assert tair_db.gdnaurl == urls, tair_db.gdnaurl
    url = ('ftp://ftp.arabidopsis.org/home/tair/Genes/TAIR6_genome_release/'
           'TAIR6_GFF3_genes.gff')
    assert tair_db.gff3url == url, tair_db.gff3url
    url = ('ftp://ftp.arabidopsis.org/home/tair/Genes/TAIR6_genome_release/'
           'TAIR6_pep_20060907')
    assert tair_db.proturl == url, tair_db.proturl


def test_annot_format():
    """TAIR6 annotation formatting"""
    conf = genhub.test_registry.genome('Att6')
    db = TairDB('Att6', conf, workdir='testdata/demo-workdir')
    db.preprocess_gff3(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Att6/Att6.gff3'
    testfile = 'testdata/gff3/tair6-format.gff3'
    assert filecmp.cmp(outfile, testfile), 'TAIR6 annotation formatting failed'


def test_protids():
    """TAIR6: extract protein IDs from GFF3"""
    conf = genhub.test_registry.genome('Att6')
    db = TairDB('Att6', conf)
    protids = ['AT5G01010.1', 'AT5G01015.1', 'AT5G01020.1', 'AT5G01030.1',
               'AT5G01030.2', 'AT5G01040.1', 'AT5G01050.1', 'AT5G01060.1',
               'AT5G01070.1', 'AT5G01075.1']
    infile = 'testdata/gff3/tair6-format.gff3'
    testids = list()
    with open(infile, 'r') as instream:
        for protid in db.gff3_protids(instream):
            testids.append(protid)
    assert sorted(protids) == sorted(testids), \
        'protein ID mismatch: %r %r' % (protids, testids)


def test_protmap():
    """TAIR6: extract protein-->iLocus mapping from GFF3"""
    conf = genhub.test_registry.genome('Att6')
    db = TairDB('Xtro', conf)
    mapping = {'AT2G01800.1': 'Att6ILC-09797',
               'AT2G01810.1': 'Att6ILC-09799',
               'AT2G01820.1': 'Att6ILC-09801',
               'AT2G01830.1': 'Att6ILC-09803',
               'AT2G01830.2': 'Att6ILC-09803',
               'AT2G01830.3': 'Att6ILC-09803'}
    infile = 'testdata/gff3/tair6-loci.gff3'
    testmap = dict()
    with open(infile, 'r') as instream:
        for protid, locid in db.protein_mapping(instream):
            testmap[protid] = locid
    assert mapping == testmap, \
        'protein mapping mismatch: %r %r' % (mapping, testmap)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Simple module for reading, writing, subsetting, and comparing sequences."""

from __future__ import print_function
import sys
try:
    from StringIO import StringIO
except ImportError:  # pragma: no cover
    from io import StringIO


def parse(data):
    """
    Load sequences in Fasta format.

    This generator function yields a tuple containing a defline and a sequence
    for each record in the Fasta data. Stolen shamelessly from
    http://stackoverflow.com/a/7655072/459780.
    """
    name, seq = None, []
    for line in data:
        line = line.rstrip()
        if line.startswith('>'):
            if name:
                yield (name, ''.join(seq))
            name, seq = line, []
        else:
            seq.append(line)
    if name:
        yield (name, ''.join(seq))


def format(seq, linewidth=70, outstream=sys.stdout):
    """Print a sequence in a readable format."""
    if linewidth == 0 or len(seq) <= linewidth:
        print(seq, file=outstream)
        return

    i = 0
    while i < len(seq):
        print(seq[i:i+linewidth], file=outstream)
        i += linewidth


def select(idstream, seqstream):
    ids = dict()
    for line in idstream:
        seqid = line.rstrip()
        ids[seqid] = True

    for defline, seq in parse(seqstream):
        seqid = defline[1:].split()[0]
        if seqid in ids:
            yield defline, seq


def compare(stream1, stream2):
    seqs1 = dict()
    seqs2 = dict()
    for defline, seq in parse(stream1):
        seqs1[defline] = seq
    for defline, seq in parse(stream2):
        seqs2[defline] = seq
    return seqs1 == seqs2


def test_parse():
    """Fasta parsing"""
    data = ('>seq1\n'
            'ACGT\n'
            '>seq2\n'
            'ACGTACGTACGTACGT\n'
            'ACGTACGTACGTACGT\n')
    seqs = {'>seq1': 'ACGT',
            '>seq2': 'ACGTACGTACGTACGTACGTACGTACGTACGT'}
    testseqs = dict()
    for defline, seq in parse(data.split('\n')):
        testseqs[defline] = seq
    assert seqs == testseqs, \
        'parsed sequence mismatch: %r %r' % (seqs, testseqs)
    seqs = []
    testseqs = list(parse(['']))
    assert seqs == testseqs, 'empty seqfile fail %r %r' % (seqs, testseqs)


def test_format_seq():
    """Sequence formatting"""
    seq = ('TCTCCCTCCA'
           'ACGCCCGAAC'
           'GTGTCTGCTC'
           'ATTTCAAGCA'
           'CACGCATGAA'
           'CGGCATCGCG'
           'CAGACGTGCG'
           'AGCGAGCGCA')

    sio = StringIO()
    format(seq, linewidth=0, outstream=sio)
    assert sio.getvalue() == seq + '\n'
    sio.close()

    sio = StringIO()
    format(seq, linewidth=40, outstream=sio)
    assert sio.getvalue() == ('TCTCCCTCCA'
                              'ACGCCCGAAC'
                              'GTGTCTGCTC'
                              'ATTTCAAGCA\n'
                              'CACGCATGAA'
                              'CGGCATCGCG'
                              'CAGACGTGCG'
                              'AGCGAGCGCA\n')
    sio.close()

    sio = StringIO()
    format(seq, linewidth=20, outstream=sio)
    assert sio.getvalue() == ('TCTCCCTCCA'
                              'ACGCCCGAAC\n'
                              'GTGTCTGCTC'
                              'ATTTCAAGCA\n'
                              'CACGCATGAA'
                              'CGGCATCGCG\n'
                              'CAGACGTGCG'
                              'AGCGAGCGCA\n')
    sio.close()


def test_select():
    """Sequence extraction"""
    data = ('>seq1\n'
            'ACGT\n'
            '>seq2\n'
            'ACGTACGTACGTACGT\n'
            'ACGTACGTACGTACGT')
    seqs = {'>seq1': 'ACGT'}
    testseqs = dict()
    for defline, seq in select(['seq1'], data.split('\n')):
        testseqs[defline] = seq
    assert seqs == testseqs, \
        'extracted sequence mismatch: %r %r' % (seqs, testseqs)


def test_compare():
    """Order-independent sequence comparison"""
    data1 = ('>seq1\n'
             'ACGT\n'
             '>seq2\n'
             'ACGTACGTACGTACGT\n'
             'ACGTACGTACGTACGT')
    data2 = ('>seq2\n'
             'ACGTACGTACGTACGT\n'
             'ACGTACGTACGTACGT\n'
             '>seq1\n'
             'ACGT')

    assert compare(data1.split('\n'), data2.split('\n')), \
        'sequence comparison failed'



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Module for handling cd-hit output."""

from __future__ import print_function
import re


class ClusterSeq(object):
    """Object representing a sequence entry in a cd-hit cluster."""

    def __init__(self, line):
        self.rawdata = line.rstrip()
        values = re.compile("\s+").split(self.rawdata)
        self.index = int(values[0])
        self.length = int(values[1][:-3])
        self.defline = values[2]

    @property
    def accession(self):
        """
        Parse accession number from the specified format.

            >gnl|Tcas|XP_008191512.1
        """
        assert self.defline.startswith('>gnl|')
        acc = re.match('>gnl\|[^\|]+\|([^\|\n]+)', self.defline).group(1)
        assert acc.endswith('...')
        acc = acc[:-3]
        return acc

    @property
    def species(self):
        """
        Parse species label from the specified format.

            >gnl|Tcas|XP_008191512.1
        """
        return self.defline.split('|')[1]

    def __len__(self):
        return self.length


def parse_clusters(filehandle):
    """
    Iterate over clusters from a CD-HIT output file.

    Yields the cluster ID (a numeric string) and a list of sequence objects.
    """
    clusterid = None
    clusterseqs = list()
    for line in filehandle:
        if line.startswith('>'):
            if clusterid is not None:
                yield clusterid, clusterseqs
            clusterid = line.rstrip()[9:]  # Strip '>Cluster ' from front
            clusterseqs = list()
        else:
            seqinfo = ClusterSeq(line)
            clusterseqs.append(seqinfo)

    yield clusterid, clusterseqs


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------

def test_parse_clusters():
    """CD-HIT: parse clusters"""
    clusters = list()
    with open('testdata/misc/hymhub-head.clstr', 'r') as infile:
        for clusterid, clusterseqs in parse_clusters(infile):
            clusters.append(clusterseqs)

    assert len(clusters) == 3

    assert len(clusters[0]) == 1
    assert len(clusters[0][0]) == 25481
    assert clusters[0][0].species == 'Tcas'
    assert clusters[0][0].accession == 'XP_008191512.1'

    assert len(clusters[1]) == 1
    assert len(clusters[1][0]) == 22949
    assert clusters[1][0].species == 'Dmel'
    assert clusters[1][0].accession == 'NP_001260032.1'

    assert len(clusters[2]) == 3
    assert clusters[2][0].species == 'Amel'
    assert clusters[2][1].species == 'Bimp'
    assert clusters[2][2].species == 'Bter'



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Genome database implementation for FlyBase data at NCBI."""

from __future__ import print_function
import filecmp
import gzip
import os
import re
import subprocess
import sys
import genhub


class FlyBaseDB(genhub.genomedb.GenomeDB):

    def __init__(self, label, conf, workdir='.'):
        super(FlyBaseDB, self).__init__(label, conf, workdir)
        assert self.config['source'] == 'ncbi_flybase'
        assert 'species' in self.config
        species = self.config['species'].replace(' ', '_')
        self.specbase = ('ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                         'Drosophila_melanogaster/RELEASE_5_48')
        self.format_gdna = self.format_fasta
        self.format_prot = self.format_fasta

    def __repr__(self):
        return 'FlyBase@NCBI'

    @property
    def gdnafilename(self):
        return '%s.orig.fa.gz' % self.label

    @property
    def gdnaurl(self):
        urls = list()
        for acc in self.config['accessions']:
            url = '%s/%s.fna' % (self.specbase, acc)
            urls.append(url)
        return urls

    @property
    def gff3url(self):
        urls = list()
        for acc in self.config['accessions']:
            url = '%s/%s.gff' % (self.specbase, acc)
            urls.append(url)
        return urls

    @property
    def proturl(self):
        urls = list()
        for acc in self.config['accessions']:
            url = '%s/%s.faa' % (self.specbase, acc)
            urls.append(url)
        return urls

    def download_gff3(self, logstream=sys.stderr):  # pragma: no cover
        """Override the default download task."""
        subprocess.call(['mkdir', '-p', self.dbdir])
        if logstream is not None:
            logmsg = '[GenHub: %s] ' % self.config['species']
            logmsg += 'download genome annotation from %r' % self
            print(logmsg, file=logstream)

        command = ['gt', 'gff3', '-sort', '-tidy', '-force', '-gzip', '-o',
                   '%s' % self.gff3path]
        for url, acc in zip(self.gff3url, self.config['accessions']):
            tempout = '%s/%s.gff.gz' % (self.dbdir, os.path.basename(acc))
            genhub.download.url_download(url, tempout, compress=True)
            command.append(tempout)
        logfile = open('%s.log' % self.gff3path, 'w')
        proc = subprocess.Popen(command, stderr=subprocess.PIPE)
        proc.wait()
        for line in proc.stderr:
            print(line, end='', file=logfile)
        assert proc.returncode == 0, ('command failed, check the log '
                                      '(%s.log): %s' %
                                      (self.gff3path, ' '.join(command)))

    def format_fasta(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            if line.startswith('>'):
                pattern = '>gi\|\d+\|(ref|gb)\|([^\|]+)\S+'
                line = re.sub(pattern, '>\g<2>', line)
            print(line, end='', file=outstream)

    def format_gff3(self, logstream=sys.stderr, debug=False):
        cmds = list()
        cmds.append('gunzip -c %s' % self.gff3path)
        excludefile = self.filter_file()
        cmds.append('grep -vf %s' % excludefile.name)
        cmds.append('genhub-fix-trna.py')
        cmds.append('tidygff3')
        cmds.append('genhub-format-gff3.py --source ncbi_flybase -')
        cmds.append('gt gff3 -sort -tidy -o %s -force' % self.gff3file)

        commands = ' | '.join(cmds)
        if debug:  # pragma: no cover
            print('DEBUG: running command: %s' % commands, file=logstream)
        proc = subprocess.Popen(commands, shell=True, universal_newlines=True,
                                stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        for line in stderr.split('\n'):  # pragma: no cover
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               line != '':
                print(line, file=logstream)
        assert proc.returncode == 0, \
            'annot cleanup command failed: %s' % commands
        os.unlink(excludefile.name)

    def gff3_protids(self, instream):
        protids = dict()
        for line in instream:
            if '\tCDS\t' not in line:
                continue
            idmatch = re.search('protein_id=([^;\n]+)', line)
            assert idmatch, 'cannot parse protein_id: ' + line
            protid = idmatch.group(1)
            if protid not in protids:
                protids[protid] = True
                yield protid

    def protein_mapping(self, instream):
        locusid2name = dict()
        gene2loci = dict()
        mrna2gene = dict()
        proteins = dict()
        for line in instream:
            fields = line.split('\t')
            if len(fields) != 9:
                continue
            feattype = fields[2]
            attrs = fields[8]

            if feattype == 'locus':
                idmatch = re.search('ID=([^;\n]+);.*Name=([^;\n]+)', attrs)
                if idmatch:
                    locusid = idmatch.group(1)
                    locusname = idmatch.group(2)
                    locusid2name[locusid] = locusname
            elif feattype == 'gene':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse gene and iLocus IDs: %s' % attrs
                geneid = idmatch.group(1)
                ilocusid = idmatch.group(2)
                gene2loci[geneid] = ilocusid
            elif feattype == 'mRNA':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse mRNA and gene IDs: %s' % attrs
                mrnaid = idmatch.group(1)
                geneid = idmatch.group(2)
                mrna2gene[mrnaid] = geneid
            elif feattype == 'CDS':
                idmatch = re.search('Parent=([^;\n]+).*protein_id=([^;\n]+)',
                                    attrs)
                assert idmatch, \
                    'Unable to parse protein and mRNA IDs: %s' % attrs
                mrnaid = idmatch.group(1)
                proteinid = idmatch.group(2)
                if proteinid not in proteins:
                    geneid = mrna2gene[mrnaid]
                    proteins[proteinid] = mrnaid
                    ilocusid = gene2loci[geneid]
                    ilocusname = locusid2name[ilocusid]
                    yield proteinid, ilocusname


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_chromosomes():
    """NCBI/FlyBase chromosome download"""
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config)

    testurls = ['ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_X/NC_004354.fna',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_2/NT_033778.fna',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_2/NT_033779.fna',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_3/NT_033777.fna',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_3/NT_037436.fna',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_4/NC_004353.fna']
    testpath = './Dmfb/Dmfb.orig.fa.gz'
    assert dmel_db.gdnaurl == testurls, \
        'chromosome URL mismatch\n%s\n%s' % (dmel_db.gdnaurl, testurls)
    assert dmel_db.gdnapath == testpath, \
        'chromosome path mismatch\n%s\n%s' % (dmel_db.gdnapath, testpath)
    assert '%r' % dmel_db == 'FlyBase@NCBI'
    assert dmel_db.compress_gdna is True


def test_annot():
    """NCBI/FlyBase annotation download"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config)

    testurls = ['ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_X/NC_004354.gff',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_2/NT_033778.gff',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_2/NT_033779.gff',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_3/NT_033777.gff',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_3/NT_037436.gff',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_4/NC_004353.gff']
    testpath = './Dmfb/dmel-5.48-ncbi.gff3.gz'
    assert dmel_db.gff3url == testurls, \
        'annotation URL mismatch\n%s\n%s' % (dmel_db.gff3url, testurls)
    assert dmel_db.gff3path == testpath, \
        'annotation path mismatch\n%s\n%s' % (dmel_db.gff3path, testpath)
    assert dmel_db.compress_gff3 is True


def test_proteins():
    """NCBI/FlyBase protein download"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config)

    testurls = ['ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_X/NC_004354.faa',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_2/NT_033778.faa',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_2/NT_033779.faa',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_3/NT_033777.faa',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_3/NT_037436.faa',
                'ftp://ftp.ncbi.nih.gov/genomes/archive/old_refseq/'
                'Drosophila_melanogaster/RELEASE_5_48/CHR_4/NC_004353.faa']
    testpath = './Dmfb/protein.fa.gz'
    assert dmel_db.proturl == testurls, \
        'protein URL mismatch\n%s\n%s' % (dmel_db.proturl, testurls)
    assert dmel_db.protpath == testpath, \
        'protein path mismatch\n%s\n%s' % (dmel_db.protpath, testpath)
    assert dmel_db.compress_prot is True


def test_format():
    """Task drivers"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config, workdir='testdata/demo-workdir')
    dmel_db.format(logstream=None, verify=False)


def test_gdna_format():
    """NCBI/FlyBase gDNA formatting"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config, workdir='testdata/demo-workdir')

    dmel_db.preprocess_gdna(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Dmfb/Dmfb.gdna.fa'
    testoutfile = 'testdata/fasta/dmel-fb-gdna-ut-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Dmfb gDNA formatting failed'


def test_annot_format():
    """NCBI/FlyBase annotation formatting"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config, workdir='testdata/demo-workdir')

    dmel_db.preprocess_gff3(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Dmfb/Dmfb.gff3'
    testfile = 'testdata/gff3/ncbi-format-dmel.gff3'
    assert filecmp.cmp(outfile, testfile), 'Dmfb annotation formatting failed'


def test_prot_format():
    """NCBI/FlyBase protein formatting"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config, workdir='testdata/demo-workdir')

    dmel_db.preprocess_prot(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Dmfb/Dmfb.all.prot.fa'
    testoutfile = 'testdata/fasta/dmel-fb-prot-ut-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Dmfb protein formatting failed'


def test_protids():
    """NCBI/FlyBase: extract protein IDs from GFF3"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config)

    protids = ['NP_524820.2', 'NP_001259789.1', 'NP_608489.2']
    infile = 'testdata/gff3/dmel-net.gff3'
    testids = list()
    with open(infile, 'r') as instream:
        for protid in dmel_db.gff3_protids(instream):
            testids.append(protid)
    assert sorted(protids) == sorted(testids), \
        'protein ID mismatch: %r %r' % (protids, testids)


def test_protmap():
    """NCBI/FlyBase: extract protein-->iLocus mapping from GFF3"""
    registry = genhub.registry.Registry()
    config = genhub.test_registry.genome('Dmfb')
    dmel_db = FlyBaseDB('Dmfb', config)

    mapping = {'NP_001259789.1': 'DmfbILC-10965',
               'NP_524820.2': 'DmfbILC-10965',
               'NP_608489.2': 'DmfbILC-10967',
               'NP_608490.1': 'DmfbILC-10968',
               'NP_001259790.1': 'DmfbILC-10968'}
    infile = 'testdata/gff3/dmel-net-loci.gff3'
    testmap = dict()
    with open(infile, 'r') as instream:
        for protid, locid in dmel_db.protein_mapping(instream):
            testmap[protid] = locid
    assert mapping == testmap, \
        'protein mapping mismatch: %r %r' % (mapping, testmap)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

from __future__ import print_function
import argparse
import re
import sys
import genhub


def gc_content(dna):
    """Calculate the %GC content of a nucleotide sequence."""
    seqlength = len(dna)

    # Count A and T nucleotides, including the W ambiguity base representing
    # either A or T
    atcount = dna.count('A') + dna.count('a') + \
        dna.count('T') + dna.count('t') + \
        dna.count('W') + dna.count('w')

    # Count C and G nucleotides, including the S ambiguity base representing
    # either C or G
    gccount = dna.count('C') + dna.count('c') + \
        dna.count('G') + dna.count('g') + \
        dna.count('S') + dna.count('s')

    # Count all other ambiguous nucleotides; most will be Ns, but occasionally
    # there will be other IUPAC ambiguity symbols
    ncount = seqlength - atcount - gccount

    if atcount + gccount == 0:
        assert ncount == seqlength
        gccontent = 0.0
    else:
        gccontent = float(gccount) / float(gccount + atcount)
    return gccontent


def gc_skew(dna):
    """
    Calculate the GC skew of a nucleotide sequence.

    s = (G - C) / (G + C)
    """
    gcount = dna.count('G') + dna.count('g')
    ccount = dna.count('C') + dna.count('c')
    if gcount + ccount == 0:
        return 0.0
    return float(gcount - ccount) / float(gcount + ccount)


def n_content(dna):
    """Calculate the proportion of ambigious nucleotides in a DNA sequence."""
    ncount = dna.count('N') + dna.count('n') + \
        dna.count('X') + dna.count('x')
    if ncount == 0:
        return 0.0
    return float(ncount) / float(len(dna))


def ilocus_desc(gff3, fasta, miloci=False):
    """
    Generate a tabular record for each iLocus in the input.

    - gff3: file handle to a GFF3 file containing iLocus annotations
    - fasta: file handle to a Fasta file containing iLocus sequences
    """
    seqs = {}
    for defline, seq in genhub.fasta.parse(fasta):
        seqid = defline[1:].split(' ')[0]
        assert seqid not in seqs
        seqs[seqid] = seq

    for entry in gff3:
        if '\tlocus\t' not in entry:
            continue
        fields = entry.rstrip().split('\t')
        assert len(fields) == 9
        locuspos = '%s_%s-%s' % (fields[0], fields[3], fields[4])
        if miloci:
            locuspos = 'locus:%s.' % locuspos
        locusid = locuspos
        locusidmatch = re.search('Name=([^;\n]+)', fields[8])
        if locusidmatch:
            locusid = locusidmatch.group(1)
        locuslen = int(fields[4]) - int(fields[3]) + 1
        locusseq = seqs[locusid]
        assert len(locusseq) == locuslen, \
            'Locus "%s": length mismatch; gff=%d, fa=%d' % (
            locusid, locuslen, len(locusseq))
        gccontent = gc_content(locusseq)
        gcskew = gc_skew(locusseq)
        ncontent = n_content(locusseq)

        classmatch = re.search('iLocus_type=([^;\n]+)', fields[8])
        assert(classmatch), fields[8]
        locusclass = classmatch.group(1)
        genecount = 0
        attrs = fields[8]
        unannot = 'unannot=true' in attrs
        efflen = 0
        efflenmatch = re.search('effective_length=(\d+)', attrs)
        if efflenmatch:
            efflen = int(efflenmatch.group(1))
        if 'gene=' in attrs:
            locustype = 'gene'
            gmatch = re.search('gene=(\d+)', attrs)
            assert gmatch
            genecount = int(gmatch.group(1))
        orient = 'NA'
        orientmatch = re.search('fg_orient=(..)', attrs)
        if orientmatch:
            orient = orientmatch.group(1)
        values = '%s %s %d %d %.3f %.3f %.3f %s %d %r %s' % (
            locusid, locuspos, locuslen, efflen, gccontent, gcskew, ncontent,
            locusclass, genecount, unannot, orient)
        yield values.split(' ')


def premrna_desc(gff3, fasta):
    """
    Given pre-mRNA sequences and corresponding annotations, generate a tabular
    record for each.
    """
    seqs = {}
    for defline, seq in genhub.fasta.parse(fasta):
        seqid = defline[1:].split(' ')[0]
        if seqid not in seqs:
            seqs[seqid] = seq

    mrnaacc = ''
    mrnalen = 0
    gccontent = 0.0
    gcskew = 0.0
    ncontent = 0.0
    exoncount = 0
    introncount = 0
    utr5plen = 0
    utr3plen = 0
    for entry in gff3:
        if '\tmRNA\t' in entry:
            fields = entry.rstrip().split('\t')
            assert len(fields) == 9
            mrnaacc = re.search('accession=([^;\n]+)', fields[8]).group(1)
            mrnalen = int(fields[4]) - int(fields[3]) + 1
            mrnaseq = seqs[mrnaacc]
            if len(mrnaseq) != mrnalen:
                message = 'pre-mRNA "%s": length mismatch' % mrnaacc
                message += ' (gff3=%d, fa=%d)' % (mrnalen, len(mrnaseq))
                message += '; most likely a duplicated accession, discarding'
                print(message, file=sys.stderr)
                mrnaacc = ''
            gccontent = gc_content(mrnaseq)
            gcskew = gc_skew(mrnaseq)
            ncontent = n_content(mrnaseq)
        elif '\texon\t' in entry:
            exoncount += 1
        elif '\tintron\t' in entry:
            introncount += 1
        elif '\tfive_prime_UTR\t' in entry:
            fields = entry.rstrip().split('\t')
            assert len(fields) == 9
            utr5plen += int(fields[4]) - int(fields[3]) + 1
        elif '\tthree_prime_UTR\t' in entry:
            fields = entry.rstrip().split('\t')
            assert len(fields) == 9
            utr3plen += int(fields[4]) - int(fields[3]) + 1
        elif entry.startswith('###'):
            if mrnaacc != '':
                values = '%s %d %.3f %.3f %.3f %d %d %d %d' % (
                    mrnaacc, mrnalen, gccontent, gcskew, ncontent,
                    exoncount, introncount, utr5plen, utr3plen)
                yield values.split(' ')
            mrnaacc = ''
            mrnalen = 0
            gccontent = 0.0
            gcskew = 0.0
            ncontent = 0.0
            exoncount = 0
            exonlen = 0
            introncount = 0
            utr5plen = 0
            utr3plen = 0


def mrna_desc(gff3, fasta):
    """
    Given mature (sans introns) mRNA sequences and their corresponding
    annotations, generate a tabular record for each mRNA.
    """
    seqs = {}
    for defline, seq in genhub.fasta.parse(fasta):
        seqid = defline[1:].split(' ')[0]
        if seqid not in seqs:
            seqs[seqid] = seq

    mrnaacc = ''
    mrnalen = 0
    for entry in gff3:
        if '\tmRNA\t' in entry:
            fields = entry.rstrip().split('\t')
            assert len(fields) == 9
            mrnalen += int(fields[4]) - int(fields[3]) + 1
            accmatch = re.search('accession=([^;\n]+)', fields[8])
            assert accmatch, 'Unable to parse mRNA accession: %s' % fields[8]
            mrnaacc = accmatch.group(1)
        elif entry.startswith('###'):
            mrnaseq = seqs[mrnaacc]
            if len(mrnaseq) != mrnalen:
                message = 'mature mRNA "%s": length mismatch' % mrnaacc
                message += ' (gff3=%d, fa=%d)' % (mrnalen, len(mrnaseq))
                message += '; most likely a duplicated accession, discarding'
                print(message, file=sys.stderr)
            else:
                gccontent = gc_content(mrnaseq)
                gcskew = gc_skew(mrnaseq)
                ncontent = n_content(mrnaseq)
                values = '%s %d %.3f %.3f %.3f' % (
                    mrnaacc, mrnalen, gccontent, gcskew, ncontent)
                yield values.split(' ')
            mrnaacc = ''
            mrnalen = 0


def cds_desc(gff3, fasta):
    """
    Given CDS sequences and their corresponding annotations, generate a tabular
    record for each CDS.
    """
    seqs = {}
    for defline, seq in genhub.fasta.parse(fasta):
        seqid = defline[1:].split(' ')[0]
        if seqid not in seqs:
            seqs[seqid] = seq

    accession = ''
    cdslen = 0
    for entry in gff3:
        if '\tCDS\t' in entry:
            fields = entry.rstrip().split('\t')
            assert len(fields) == 9
            accession = re.search('accession=([^;\n]+)', fields[8]).group(1)
            cdslen += int(fields[4]) - int(fields[3]) + 1
        elif entry.startswith('###'):
            if accession:
                cdsseq = seqs[accession]
                if len(cdsseq) != cdslen:
                    message = 'CDS for "%s": length mismatch' % accession
                    message += ' (gff3=%d, fa=%d)' % (cdslen, len(cdsseq))
                    message += '; most likely a duplicated accession'
                    message += ', discarding'
                    print(message, file=sys.stderr)
                else:
                    gccontent = gc_content(cdsseq)
                    gcskew = gc_skew(cdsseq)
                    ncontent = n_content(cdsseq)
                    values = '%s %d %.3f %.3f %.3f' % (
                        accession, cdslen, gccontent, gcskew, ncontent)
                    yield values.split(' ')
            accession = ''
            cdslen = 0


def feat_overlap(f1, f2):
    """
    Given two features (lists of length=9 from GFF3), determine whether they
    overlap.
    """
    f1start = int(f1[3])
    f1end = int(f1[4])
    f2start = int(f2[3])
    f2end = int(f2[4])

    if f1start <= f2end and f1end >= f2start:
        return True
    return False


def exon_context(exon, start, stop):
    """
    Given an exon, a start codon, and a stop codon (GFF3 entries),
    determine the context of the exon:
      - cds (entirely coding)
      - 5putr (entirely 5' UTR)
      - 3putr (entirely 3' UTR)
      - start (includes start codon)
      - stop (includes stop codon)
      - complete (includes both start and stop codon)
    """
    assert start and stop
    exon = exon.split('\t')
    start = start.split('\t')
    stop = stop.split('\t')
    assert len(exon) == 9 and len(start) == 9 and len(stop) == 9

    hasstart = feat_overlap(exon, start)
    hasstop = feat_overlap(exon, stop)
    if hasstart or hasstop:
        if hasstart and hasstop:
            return 'complete'
        elif hasstart:
            return 'start'
        else:
            assert hasstop
            return 'stop'

    exonstart = int(exon[3])
    exonend = int(exon[4])
    codonnucs = [start[3], start[4], stop[3], stop[4]]
    codonnucs = [int(x) for x in codonnucs]
    leftmostnuc = min(codonnucs)
    rightmostnuc = max(codonnucs)
    if exonend < leftmostnuc:
        if exon[6] == '-':
            return '3putr'
        else:
            return '5putr'
    elif exonstart > rightmostnuc:
        if exon[6] == '-':
            return '5putr'
        else:
            return '3putr'
    else:
        assert exonstart > leftmostnuc and exonend < rightmostnuc
        return 'cds'


def exon_desc(gff3, fasta):
    """
    Given exon sequences and their corresponding annotations, generate a
    tabular record for each exon.
    """
    seqs = {}
    for defline, seq in genhub.fasta.parse(fasta):
        exonpos = defline[1:].split(' ')[1]
        seqs[exonpos] = seq

    rnaid_to_accession = dict()
    reported_exons = {}
    exons, cdss = [], {}
    start, stop = None, None
    moltypes = ['mRNA', 'tRNA', 'ncRNA', 'transcript', 'primary_transcript',
                'V_gene_segment', 'D_gene_segment', 'J_gene_segment',
                'C_gene_segment']
    for entry in gff3:
        for moltype in moltypes:
            if ('\t%s\t' % moltype) in entry:
                accession = re.search('accession=([^;\n]+)', entry).group(1)
                tid = re.search('ID=([^;\n]+)', entry).group(1)
                rnaid_to_accession[tid] = accession

        if '\texon\t' in entry:
            exons.append(entry)
        elif '\tCDS\t' in entry:
            fields = entry.split('\t')
            pos = '%s_%s-%s%s' % (fields[0], fields[3], fields[4], fields[6])
            cdss[pos] = entry
        elif '\tstart_codon\t' in entry:
            start = entry
        elif '\tstop_codon\t' in entry:
            stop = entry
        elif entry.startswith('###'):
            if len(exons) == 0:
                continue
            xcept = False
            for exonpos in cdss:
                if ';exception=ribosomal slippage' in cdss[exonpos]:
                    xcept = True
            if xcept:
                exons, cdss = [], {}
                start, stop = None, None
                continue
            assert start, 'No start codon for exon(s): %s' % exons[0]
            assert stop,  'No stop codon for exon(s): %s' % exons[0]
            for exon in exons:
                fields = exon.split('\t')
                assert len(
                    fields) == 9, 'entry does not have 9 fields: %s' % exon
                mrnaid = re.search('Parent=([^;\n]+)', fields[8]).group(1)
                exonpos = '%s_%s-%s%s' % (fields[0],
                                          fields[3], fields[4], fields[6])
                if exonpos in reported_exons:
                    continue
                exonlength = int(fields[4]) - int(fields[3]) + 1
                exonseq = seqs[exonpos]
                assert len(exonseq) == exonlength, \
                    'exon "%s": length mismatch; gff=%d, fa=%d' % (
                    exonpos, exonlength, len(exonseq))
                gccontent = gc_content(exonseq)
                gcskew = gc_skew(exonseq)
                ncontent = n_content(exonseq)
                context = exon_context(exon, start, stop)
                phase = None
                remainder = None
                if context == 'cds':
                    cexon = cdss[exonpos]
                    phase = int(cexon.split('\t')[7])
                    remainder = (exonlength - phase) % 3
                values = '%s %s %d %.3f %.3f %.3f %s %r %r' % (
                    exonpos, rnaid_to_accession[mrnaid], exonlength, gccontent,
                    gcskew, ncontent, context, phase, remainder)
                reported_exons[exonpos] = 1
                yield values.split(' ')
            exons, cdss = [], {}
            start, stop = None, None


def intron_context(intron, start, stop):
    """
    Given an intron, a start codon, and a stop codon (GFF3 entries),
    determine the context of the exon:
      - cds (entirely coding)
      - 5putr (entirely 5' UTR)
      - 3putr (entirely 3' UTR)
      - start (includes start codon)
      - stop (includes stop codon)
      - complete (includes both start and stop codon)
    """
    assert start and stop
    intron = intron.split('\t')
    start = start.split('\t')
    stop = stop.split('\t')
    assert len(intron) == 9 and len(start) == 9 and len(stop) == 9

    intronstart = int(intron[3])
    intronend = int(intron[4])
    codonnucs = [start[3], start[4], stop[3], stop[4]]
    codonnucs = [int(x) for x in codonnucs]
    leftmostnuc = min(codonnucs)
    rightmostnuc = max(codonnucs)
    if intronend < leftmostnuc:
        if intron[6] == '-':
            return '3putr'
        else:
            return '5putr'
    elif intronstart > rightmostnuc:
        if intron[6] == '-':
            return '5putr'
        else:
            return '3putr'
    else:
        assert intronstart > leftmostnuc and intronend < rightmostnuc
        return 'cds'


def intron_desc(gff3, fasta):
    """
    Given intron sequences and their corresponding annotations, generate a
    tabular record for each intron.
    """
    seqs = {}
    for defline, seq in genhub.fasta.parse(fasta):
        intronpos = defline[1:].split(' ')[1]
        seqs[intronpos] = seq

    reported_introns = {}
    introns = []
    mrnaid = None
    start, stop = None, None
    for entry in gff3:
        if '\tmRNA\t' in entry:
            mrnaid = re.search('accession=([^;\n]+)', entry).group(1)
        elif '\tintron\t' in entry:
            introns.append(entry)
        elif '\tstart_codon\t' in entry:
            start = entry
        elif '\tstop_codon\t' in entry:
            stop = entry
        elif entry.startswith('###'):
            if mrnaid is None:
                continue
            assert start, 'No start codon for introns(s): %s' % introns[0]
            assert stop,  'No stop codon for introns(s): %s' % introns[0]
            if len(introns) > 0:
                for intron in introns:
                    fields = intron.split('\t')
                    assert len(fields) == 9, \
                        'entry does not have 9 fields: %s' % intron
                    intronpos = '%s_%s-%s%s' % (fields[0], fields[3],
                                                fields[4], fields[6])
                    if intronpos in reported_introns:
                        continue
                    intronlength = int(fields[4]) - int(fields[3]) + 1
                    intronseq = seqs[intronpos]
                    assert len(intronseq) == intronlength, \
                        'intron "%s": length mismatch; gff=%d, fa=%d' % (
                            intronpos, intronlength, len(intronseq))
                    gccontent = gc_content(intronseq)
                    gcskew = gc_skew(intronseq)
                    ncontent = n_content(intronseq)
                    context = intron_context(intron, start, stop)
                    values = '%s %s %d %.3f %.3f %.3f %s' % (
                        intronpos, mrnaid, intronlength, gccontent, gcskew,
                        ncontent, context)
                    reported_introns[intronpos] = 1
                    yield values.split(' ')
            mrnaid = None
            introns = []
            start, stop = None, None

if __name__ == '__main__':
    desc = 'Calculate descriptive statistics of genome features'
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('--species', type=str, default='default',
                        metavar='Spec', help='specify species label')
    parser.add_argument('--iloci', type=str, nargs=3,
                        metavar=('gff', 'fa', 'out'),
                        help='compute iLocus statistics')
    parser.add_argument('--miloci', type=str, nargs=3,
                        metavar=('gff', 'fa', 'out'),
                        help='compute merged iLocus statistics')
    parser.add_argument('--prnas', type=str, nargs=3,
                        metavar=('gff', 'fa', 'out'),
                        help='compute stats on pre-mRNAs')
    parser.add_argument('--mrnas', type=str, nargs=3,
                        metavar=('gff', 'fa', 'out'),
                        help='compute stats on mature mRNAs')
    parser.add_argument('--cds', type=str, nargs=3,
                        metavar=('gff', 'fa', 'out'),
                        help='compute CDS statistics')
    parser.add_argument('--exons', type=str, nargs=3,
                        metavar=('gff', 'fa', 'out'),
                        help='compute exon statistics')
    parser.add_argument('--introns', type=str, nargs=3,
                        metavar=('gff', 'fa', 'out'),
                        help='compute intron statistics')
    args = parser.parse_args()

    # Process iLoci
    if args.iloci:
        a = args.iloci
        with open(a[0], 'r') as gff, \
                open(a[1], 'r') as fa,  \
                open(a[2], 'w') as out:
            header = ['Species', 'LocusId', 'LocusPos', 'Length',
                      'EffectiveLength', 'GCContent', 'GCSkew', 'NContent',
                      'LocusClass', 'GeneCount', 'SeqUnannot',
                      'FlankGeneOrient']
            print('\t'.join(header), file=out)
            for fields in ilocus_desc(gff, fa):
                fields = [args.species] + fields
                print('\t'.join(fields), file=out)

    # Process miLoci
    if args.miloci:
        a = args.miloci
        with open(a[0], 'r') as gff, \
                open(a[1], 'r') as fa,  \
                open(a[2], 'w') as out:
            header = ['Species', 'LocusId', 'LocusPos', 'Length',
                      'EffectiveLength', 'GCContent', 'GCSkew', 'NContent',
                      'LocusClass', 'GeneCount', 'SeqUnannot',
                      'FlankGeneOrient']
            print('\t'.join(header), file=out)
            for fields in ilocus_desc(gff, fa, miloci=True):
                fields = [args.species] + fields
                print('\t'.join(fields), file=out)

    # Process pre-mRNAs
    if args.prnas:
        a = args.prnas
        with open(a[0], 'r') as gff, \
                open(a[1], 'r') as fa,  \
                open(a[2], 'w') as out:
            header = ['Species', 'Accession', 'Length', 'GCContent',
                      'GCSkew', 'NContent', 'ExonCount', 'IntronCount',
                      '5pUTRlen', '3pUTRlen']
            print('\t'.join(header), file=out)
            for fields in premrna_desc(gff, fa):
                fields = [args.species] + fields
                print('\t'.join(fields), file=out)

    # Process mature mRNAs
    if args.mrnas:
        a = args.mrnas
        with open(a[0], 'r') as gff, \
                open(a[1], 'r') as fa, \
                open(a[2], 'w') as out:
            header = ['Species', 'Accession', 'Length', 'GCContent',
                      'GCSkew', 'NContent']
            print('\t'.join(header), file=out)
            for fields in mrna_desc(gff, fa):
                fields = [args.species] + fields
                print('\t'.join(fields), file=out)

    # Process coding sequences
    if args.cds:
        a = args.cds
        with open(a[0], 'r') as gff, \
                open(a[1], 'r') as fa, \
                open(a[2], 'w') as out:
            header = ['Species', 'MrnaAcc', 'Length', 'GCContent',
                      'GCSkew', 'NContent']
            print('\t'.join(header), file=out)
            for fields in cds_desc(gff, fa):
                fields = [args.species] + fields
                print('\t'.join(fields), file=out)

    # Process exons
    if args.exons:
        a = args.exons
        with open(a[0], 'r') as gff, \
                open(a[1], 'r') as fa, \
                open(a[2], 'w') as out:
            header = ['Species', 'ExonPos', 'MrnaAcc', 'Length', 'GCContent',
                      'GCSkew', 'NContent', 'Context', 'Phase', 'Remainder']
            print('\t'.join(header), file=out)
            for fields in exon_desc(gff, fa):
                fields = [args.species] + fields
                print('\t'.join(fields), file=out)

    # Process introns
    if args.introns:
        a = args.introns
        with open(a[0], 'r') as gff, \
                open(a[1], 'r') as fa, \
                open(a[2], 'w') as out:
            header = ['Species', 'IntronPos', 'MrnaAcc', 'Length', 'GCContent',
                      'GCSkew', 'NContent', 'Context']
            print('\t'.join(header), file=out)
            for fields in intron_desc(gff, fa):
                fields = [args.species] + fields
                print('\t'.join(fields), file=out)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Package-wide configuration"""

# Package modules
from __future__ import print_function
from . import registry
from . import download
from . import fasta
from . import cdhit
from . import genomedb
from . import refseq
from . import ncbi_flybase
from . import beebase
from . import crg
from . import tair
from . import iloci
from . import proteins
from . import mrnas
from . import exons
from . import stats
try:
    FileNotFoundError
except NameError:  # pragma: no cover
    FileNotFoundError = IOError

# Custom modules
from . import pdom
from . import am10

# Versioneer
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


# Unit test fixtures (can't figure out how to do package-scope global fixtures
# with nose's setup and teardown mechanism).
test_registry = registry.Registry()
test_registry_supp = registry.Registry()
try:
    # This will only work when the current working directory is the GenHub
    # root directory. Fine since it's only for development.
    test_registry_supp.update('testdata/conf')
except FileNotFoundError:  # pragma: no cover
    pass



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""Setup configuration for genhub"""

import setuptools
import versioneer


setuptools.setup(name='genhub',
                 version=versioneer.get_version(),
                 cmdclass=versioneer.get_cmdclass(),
                 description=('Collect and pre-process annotated genome '
                              'assemblies'),
                 url='http://github.com/standage/genhub',
                 author='Daniel Standage',
                 author_email='daniel.standage@gmail.com',
                 license='BSD-3',
                 packages=['genhub'],
                 scripts=['scripts/genhub-build.py',
                          'scripts/genhub-filens.py',
                          'scripts/genhub-fix-trna.py',
                          'scripts/genhub-format-gff3.py',
                          'scripts/genhub-glean-to-gff3.py',
                          'scripts/genhub-namedup.py',
                          'scripts/genhub-stats.py'],
                 install_requires=['pyyaml', 'pycurl'],
                 package_data={'genhub': ['genomes/*.yml', 'genomes/*.txt']},
                 classifiers=[
                    'Development Status :: 4 - Beta',
                    'Environment :: Console',
                    'License :: OSI Approved :: BSD License',
                    'Programming Language :: Python :: 2.7',
                    'Programming Language :: Python :: 3.3',
                    'Programming Language :: Python :: 3.4',
                    'Programming Language :: Python :: 3.5',
                    'Topic :: Scientific/Engineering :: Bio-Informatics'
                 ],
                 zip_safe=False)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""
Retrieve and format data from RefSeq.

GenomeDB implementation for data residing in NCBI's RefSeq database.
"""

from __future__ import print_function
import filecmp
import gzip
import os
import re
import subprocess
import sys
import genhub


class RefSeqDB(genhub.genomedb.GenomeDB):

    def __init__(self, label, conf, workdir='.'):
        super(RefSeqDB, self).__init__(label, conf, workdir)

        assert self.config['source'] == 'refseq'
        assert 'branch' in self.config
        assert 'species' in self.config
        assert 'accession' in self.config
        assert 'build' in self.config

        species = self.config['species'].replace(' ', '_')
        self.acc = self.config['accession'] + '_' + self.config['build']

        base = 'ftp://ftp.ncbi.nlm.nih.gov/genomes/refseq'
        url_parts = [base, self.config['branch'], species,
                     'all_assembly_versions', self.acc]
        self.specbase = '/'.join(url_parts + [self.acc])
        self.format_gdna = self.format_fasta
        self.format_prot = self.format_fasta

    def __repr__(self):
        return 'RefSeq'

    @property
    def gdnafilename(self):
        return '%s_genomic.fna.gz' % self.acc

    @property
    def gff3filename(self):
        return '%s_genomic.gff.gz' % self.acc

    @property
    def protfilename(self):
        return '%s_protein.faa.gz' % self.acc

    @property
    def gdnaurl(self):
        return '%s_genomic.fna.gz' % self.specbase

    @property
    def gff3url(self):
        return '%s_genomic.gff.gz' % self.specbase

    @property
    def proturl(self):
        return '%s_protein.faa.gz' % self.specbase

    def format_fasta(self, instream, outstream, logstream=sys.stderr):
        for line in instream:
            # No processing required since new RefSeq migration (2015-11-30).
            # If any is ever needed again, do it here.
            print(line, end='', file=outstream)

    def format_gff3(self, logstream=sys.stderr, debug=False):
        cmds = list()
        cmds.append('gunzip -c %s' % self.gff3path)
        if 'annotfilter' in self.config:
            excludefile = self.filter_file()
            cmds.append('grep -vf %s' % excludefile.name)
        cmds.append('tidygff3')
        cmds.append('genhub-format-gff3.py --source refseq -')
        if 'fixseqreg' in self.config and self.config['fixseqreg'] is True:
            cmds.append('seq-reg.py - %s' % self.gdnafile)  # pragma: no cover
        cmds.append('gt gff3 -sort -tidy -o %s -force' % self.gff3file)

        commands = ' | '.join(cmds)
        if debug:  # pragma: no cover
            print('DEBUG: running command: %s' % commands, file=logstream)
        proc = subprocess.Popen(commands, shell=True, universal_newlines=True,
                                stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        for line in stderr.split('\n'):  # pragma: no cover
            if 'has not been previously introduced' not in line and \
               'does not begin with "##gff-version"' not in line and \
               'more than one pseudogene attribute' not in line and \
               line != '':
                print(line, file=logstream)
        assert proc.returncode == 0, \
            'annot cleanup command failed: %s' % commands
        if 'annotfilter' in self.config:
            os.unlink(excludefile.name)

    def gff3_protids(self, instream):
        protids = dict()
        for line in instream:
            if '\tCDS\t' not in line:
                continue
            idmatch = re.search('protein_id=([^;\n]+)', line)
            assert idmatch, 'cannot parse protein_id: ' + line
            protid = idmatch.group(1)
            if protid not in protids:
                protids[protid] = True
                yield protid

    def protein_mapping(self, instream):
        locusid2name = dict()
        gene2loci = dict()
        mrna2gene = dict()
        proteins = dict()
        protdups = dict()
        for line in instream:
            fields = line.split('\t')
            if len(fields) != 9:
                continue
            feattype = fields[2]
            attrs = fields[8]

            if feattype == 'locus':
                idmatch = re.search('ID=([^;\n]+);.*Name=([^;\n]+)', attrs)
                if idmatch:
                    locusid = idmatch.group(1)
                    locusname = idmatch.group(2)
                    locusid2name[locusid] = locusname
            elif feattype == 'gene':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                if idmatch:
                    geneid = idmatch.group(1)
                    ilocusid = idmatch.group(2)
                    gene2loci[geneid] = ilocusid
                else:  # pragma: no cover
                    print('Unable to parse gene and iLocus IDs: %s' % attrs,
                          file=sys.stderr)
            elif feattype == 'mRNA':
                idmatch = re.search('ID=([^;\n]+);Parent=([^;\n]+)', attrs)
                assert idmatch, \
                    'Unable to parse mRNA and gene IDs: %s' % attrs.rstrip()
                mrnaid = idmatch.group(1)
                geneid = idmatch.group(2)
                mrna2gene[mrnaid] = geneid
            elif feattype == 'CDS':
                if 'exception=rearrangement required for product' in attrs:
                    continue
                idmatch = re.search('Parent=([^;\n]+).*protein_id=([^;\n]+)',
                                    attrs)
                assert idmatch, \
                    'Unable to parse protein and mRNA IDs: %s' % attrs
                mrnaid = idmatch.group(1)
                proteinid = idmatch.group(2)
                if proteinid not in proteins:
                    geneid = mrna2gene[mrnaid]
                    proteins[proteinid] = mrnaid
                    ilocusid = gene2loci[geneid]
                    ilocusname = locusid2name[ilocusid]
                    yield proteinid, ilocusname


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------


def test_genome_download():
    """RefSeq chromosomes/scaffolds download"""
    config = genhub.test_registry.genome('Ador')
    testurl = ('ftp://ftp.ncbi.nlm.nih.gov/genomes/refseq/invertebrate/'
               'Apis_dorsata/all_assembly_versions/'
               'GCF_000469605.1_Apis_dorsata_1.3/'
               'GCF_000469605.1_Apis_dorsata_1.3_genomic.fna.gz')
    testpath = './Ador/GCF_000469605.1_Apis_dorsata_1.3_genomic.fna.gz'
    ador_db = RefSeqDB('Ador', config)
    assert '%r' % ador_db == 'RefSeq'
    assert ador_db.gdnaurl == testurl, \
        'scaffold URL mismatch\n%s\n%s' % (ador_db.gdnaurl, testurl)
    assert ador_db.gdnapath == testpath, \
        'scaffold path mismatch\n%s\n%s' % (ador_db.gdnapath, testpath)
    assert ador_db.compress_gdna is False

    config = genhub.test_registry.genome('Amel')
    testurl = ('ftp://ftp.ncbi.nlm.nih.gov/genomes/refseq/invertebrate/'
               'Apis_mellifera/all_assembly_versions/'
               'GCF_000002195.4_Amel_4.5/'
               'GCF_000002195.4_Amel_4.5_genomic.fna.gz')
    testpath = './Amel/GCF_000002195.4_Amel_4.5_genomic.fna.gz'
    amel_db = RefSeqDB('Amel', config)
    assert amel_db.gdnaurl == testurl, \
        'chromosome URL mismatch\n%s\n%s' % (amel_db.gdnaurl, testurl)
    assert amel_db.gdnapath == testpath, \
        'chromosome path mismatch\n%s\n%s' % (amel_db.gdnapath, testpath)
    assert ador_db.compress_gdna is False


def test_annot_download():
    """RefSeq annotation download"""
    config = genhub.test_registry.genome('Ador')
    testurl = ('ftp://ftp.ncbi.nlm.nih.gov/genomes/refseq/invertebrate/'
               'Apis_dorsata/all_assembly_versions/'
               'GCF_000469605.1_Apis_dorsata_1.3/'
               'GCF_000469605.1_Apis_dorsata_1.3_genomic.gff.gz')
    testpath = './Ador/GCF_000469605.1_Apis_dorsata_1.3_genomic.gff.gz'
    ador_db = RefSeqDB('Ador', config)
    assert ador_db.gff3url == testurl, \
        'annotation URL mismatch\n%s\n%s' % (ador_db.gff3url, testurl)
    assert ador_db.gff3path == testpath, \
        'annotation path mismatch\n%s\n%s' % (ador_db.gff3path, testpath)
    assert ador_db.compress_gff3 is False


def test_proteins_download():
    """RefSeq protein download"""
    config = genhub.test_registry.genome('Ador')
    testurl = ('ftp://ftp.ncbi.nlm.nih.gov/genomes/refseq/invertebrate/'
               'Apis_dorsata/all_assembly_versions/'
               'GCF_000469605.1_Apis_dorsata_1.3/'
               'GCF_000469605.1_Apis_dorsata_1.3_protein.faa.gz')
    testpath = ('/home/gandalf/HymHub/Ador/'
                'GCF_000469605.1_Apis_dorsata_1.3_protein.faa.gz')
    ador_db = RefSeqDB('Ador', config, workdir='/home/gandalf/HymHub')
    assert ador_db.proturl == testurl, \
        'protein URL mismatch\n%s\n%s' % (ador_db.proturl, testurl)
    assert ador_db.protpath == testpath, \
        'protein path mismatch\n%s\n%s' % (ador_db.protpath, testpath)
    assert ador_db.compress_prot is False


def test_gdna_format():
    """RefSeq gDNA formatting"""
    conf = genhub.test_registry.genome('Hsal')
    hsal_db = RefSeqDB('Hsal', conf, workdir='testdata/demo-workdir')
    hsal_db.preprocess_gdna(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Hsal/Hsal.gdna.fa'
    testoutfile = 'testdata/fasta/hsal-first-7-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Hsal gDNA formatting failed'

    conf = genhub.test_registry.genome('Tcas')
    tcas_db = RefSeqDB('Tcas', conf, workdir='testdata/demo-workdir')
    tcas_db.preprocess_gdna(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Tcas/Tcas.gdna.fa'
    testoutfile = 'testdata/fasta/tcas-first-33-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Tcas gDNA formatting failed'


def test_annot_format():
    """RefSeq annotation formatting"""
    conf = genhub.test_registry.genome('Aech')
    aech_db = RefSeqDB('Aech', conf, workdir='testdata/demo-workdir')
    aech_db.preprocess_gff3(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Aech/Aech.gff3'
    testfile = 'testdata/gff3/ncbi-format-aech.gff3'
    assert filecmp.cmp(outfile, testfile), 'Aech annotation formatting failed'

    conf = genhub.test_registry.genome('Pbar')
    conf['annotfilter'] = 'NW_011933506.1'
    pbar_db = RefSeqDB('Pbar', conf, workdir='testdata/demo-workdir')
    pbar_db.preprocess_gff3(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Pbar/Pbar.gff3'
    testfile = 'testdata/gff3/ncbi-format-pbar.gff3'
    assert filecmp.cmp(outfile, testfile), 'Pbar annotation formatting failed'

    conf = genhub.test_registry.genome('Ador')
    conf['annotfilter'] = ['NW_006264094.1', 'NW_006263516.1']
    ador_db = RefSeqDB('Ador', conf, workdir='testdata/demo-workdir')
    ador_db.preprocess_gff3(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Ador/Ador.gff3'
    testfile = 'testdata/gff3/ncbi-format-ador.gff3'
    assert filecmp.cmp(outfile, testfile), 'Ador annotation formatting failed'


def test_prot_ncbi():
    """RefSeq protein formatting"""
    conf = genhub.test_registry.genome('Hsal')
    hsal_db = RefSeqDB('Hsal', conf, workdir='testdata/demo-workdir')
    hsal_db.preprocess_prot(logstream=None, verify=False)
    outfile = 'testdata/demo-workdir/Hsal/Hsal.all.prot.fa'
    testoutfile = 'testdata/fasta/hsal-13-prot-out.fa'
    assert filecmp.cmp(testoutfile, outfile), 'Hsal protein formatting failed'


def test_protids():
    """RefSeq: extract protein IDs from GFF3"""
    conf = genhub.test_registry.genome('Xtro')
    db = RefSeqDB('Xtro', conf)
    protids = ['XP_012809995.1', 'XP_012809996.1', 'XP_012809997.1',
               'XP_012809998.1']
    infile = 'testdata/gff3/xtro-3genes.gff3'
    testids = list()
    with open(infile, 'r') as instream:
        for protid in db.gff3_protids(instream):
            testids.append(protid)
    assert sorted(protids) == sorted(testids), \
        'protein ID mismatch: %r %r' % (protids, testids)


def test_protmap():
    """RefSeq: extract protein-->iLocus mapping from GFF3"""
    conf = genhub.test_registry.genome('Xtro')
    db = RefSeqDB('Xtro', conf)
    mapping = {'XP_012809997.1': 'XtroILC-43374',
               'XP_012809996.1': 'XtroILC-43374',
               'XP_012809995.1': 'XtroILC-43373',
               'XP_012809998.1': 'XtroILC-43374'}
    infile = 'testdata/gff3/xtro-3genes-loci.gff3'
    testmap = dict()
    with open(infile, 'r') as instream:
        for protid, locid in db.protein_mapping(instream):
            testmap[protid] = locid
    assert mapping == testmap, \
        'protein mapping mismatch: %r %r' % (mapping, testmap)


def test_cleanup():
    """RefSeq cleanup task"""
    config = genhub.test_registry.genome('Aech')
    db = RefSeqDB('Aech', config, workdir='testdata/demo-workdir')
    delfiles = ['testdata/demo-workdir/Aech/Aech.gff3']
    testfiles = db.cleanup(None, False, True)
    assert testfiles == delfiles, '%r %r' % (testfiles, delfiles)
    delfiles = ['testdata/demo-workdir/Aech/Aech.gff3',
                'testdata/demo-workdir/Aech/GCF_000204515.1_Aech_3.9_genomic.'
                'gff.gz']
    testfiles = db.cleanup(None, True, True)
    assert set(testfiles) == set(delfiles), '%r %r' % (testfiles, delfiles)

    config = genhub.test_registry.genome('Bdis')
    db = RefSeqDB('Bdis', config, workdir='testdata/demo-workdir')
    delfiles = ['testdata/demo-workdir/Bdis/Bdis.gdna.fa',
                'testdata/demo-workdir/Bdis/Bdis.gff3',
                'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.gff3',
                'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.txt',
                'testdata/demo-workdir/Bdis/Bdis.miloci.fa',
                'testdata/demo-workdir/Bdis/Bdis.miloci.gff3',
                'testdata/demo-workdir/Bdis/Bdis.mrnas.txt',
                'testdata/demo-workdir/Bdis/Bdis.simple-iloci.txt',
                'testdata/demo-workdir/Bdis/ilens.temp']
    testfiles = db.cleanup(None, False, True)
    assert set(testfiles) == set(delfiles), '%r %r' % (testfiles, delfiles)
    delfiles = ['testdata/demo-workdir/Bdis/Bdis.gdna.fa',
                'testdata/demo-workdir/Bdis/Bdis.gff3',
                'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.gff3',
                'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.txt',
                'testdata/demo-workdir/Bdis/Bdis.mrnas.txt',
                'testdata/demo-workdir/Bdis/Bdis.simple-iloci.txt',
                'testdata/demo-workdir/Bdis/ilens.temp']
    testfiles = db.cleanup(['.miloci.'], False, True)
    assert set(testfiles) == set(delfiles), '%r %r' % (testfiles, delfiles)
    delfiles = ['testdata/demo-workdir/Bdis/Bdis.gdna.fa',
                'testdata/demo-workdir/Bdis/Bdis.gff3',
                'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.gff3',
                'testdata/demo-workdir/Bdis/Bdis.ilocus.mrnas.txt',
                'testdata/demo-workdir/Bdis/Bdis.mrnas.txt',
                'testdata/demo-workdir/Bdis/ilens.temp']
    testfiles = db.cleanup(['.miloci.', 'simple'], False, True)
    assert set(testfiles) == set(delfiles), '%r %r' % (testfiles, delfiles)

    config = genhub.test_registry.genome('Vcar')
    db = RefSeqDB('Vcar', config, workdir='testdata/demo-workdir')
    nodelfile = 'testdata/demo-workdir/Vcar/Vcar.protein2ilocus.txt'
    testfiles = db.cleanup(None, False, True)
    assert nodelfile not in testfiles, 'incorrectly deleted file'


def test_get_map():
    """RefSeq get prot map"""
    config = genhub.test_registry.genome('Vcar')
    db = RefSeqDB('Vcar', config, workdir='testdata/demo-workdir')
    mapping = dict()
    for protid, locid in db.get_prot_map():
        mapping[protid] = locid
    testmap = {'XP_002945607.1': 'VcarILC-00002',
               'XP_002945976.1': 'VcarILC-00003',
               'XP_002945608.1': 'VcarILC-00004',
               'XP_002945609.1': 'VcarILC-00005'}
    assert mapping == testmap, 'get prot map fail: %r %r' % (testmap, mapping)



#!/usr/bin/env python
#
# -----------------------------------------------------------------------------
# Copyright (c) 2015-2016   Daniel Standage <daniel.standage@gmail.com>
# Copyright (c) 2015-2016   Indiana University
#
# This file is part of genhub (http://github.com/standage/genhub) and is
# licensed under the BSD 3-clause license: see LICENSE.txt.
# -----------------------------------------------------------------------------

"""
Class for managing a genome database.

By "genome database" we mean a collection of sequence, annotation, and
ancillary data files for a annotated genome assembly.

This "superclass" defines many default characteristics and behaviors that are
shared across different genome databases. Subclasses implement additional
specifics for managing data from a particular source.
"""

from __future__ import print_function
import glob
import gzip
import hashlib
import os
import subprocess
import sys
import tempfile
import genhub
try:
    FileNotFoundError
except NameError:  # pragma: no cover
    FileNotFoundError = IOError


class GenomeDB(object):

    def __init__(self, label, conf, workdir='.'):
        self.label = label
        self.config = conf
        self.workdir = workdir
        assert 'source' in conf, 'data source unconfigured'

    # ----------
    # Filenames for unprocessed data from the primary source.
    # ----------

    def file_path(self, filename, check=False, message=None):
        """
        Resolve a file's complete path, optionally checking if the file exists.
        """
        filepath = '%s/%s/%s' % (self.workdir, self.label, filename)
        if check:  # pragma: no cover
            if not os.path.exists(filepath):
                msg = 'file "%s" not found' % filepath
                if message is not None:
                    msg += '; %s' % message
                raise FileNotFoundError(msg)
        return filepath

    @property
    def gdnafilename(self):
        if 'scaffolds' in self.config:
            return self.config['scaffolds']
        return None  # pragma: no cover

    @property
    def gff3filename(self):
        return self.config['annotation']

    @property
    def protfilename(self):
        if 'proteins' in self.config:
            return self.config['proteins']
        return 'protein.fa.gz'

    # ----------
    # Complete file paths for unprocessed data.
    # ----------

    @property
    def gdnapath(self):
        return self.file_path(self.gdnafilename)

    @property
    def gff3path(self):
        return self.file_path(self.gff3filename)

    @property
    def protpath(self):
        return self.file_path(self.protfilename)

    # ----------
    # File paths for processed data.
    # ----------

    @property
    def gdnafile(self):
        filename = '%s.gdna.fa' % self.label
        return self.file_path(filename)

    @property
    def gff3file(self):
        filename = '%s.gff3' % self.label
        return self.file_path(filename)

    @property
    def protfile(self):
        filename = '%s.all.prot.fa' % self.label
        return self.file_path(filename)

    # ----------
    # Determine whether raw data files need to be compressed during download.
    # ----------

    @property
    def compress_gdna(self):
        if 'compress' in self.config and 'gdna' in self.config['compress']:
            return True
        return False

    @property
    def compress_gff3(self):
        if 'compress' in self.config and 'gff3' in self.config['compress']:
            return True
        return False

    @property
    def compress_prot(self):
        if 'compress' in self.config and 'prot' in self.config['compress']:
            return True
        return False

    # ----------
    # Miscellaneous properties.
    # ----------

    @property
    def source(self):
        """The institutional source of the data."""
        return self.config['source']

    @property
    def dbdir(self):
        """Dedicated directory for this genome database."""
        return '%s/%s' % (self.workdir, self.label)

    # ----------
    # Build task method implementations.
    # ----------

    def download_gdna(self, logstream=sys.stderr):  # pragma: no cover
        """Download genomic DNA sequence."""
        subprocess.call(['mkdir', '-p', self.dbdir])
        if logstream is not None:
            logmsg = '[GenHub: %s] ' % self.config['species']
            logmsg += 'download genome sequence from %r' % self
            print(logmsg, file=logstream)
        genhub.download.url_download(self.gdnaurl, self.gdnapath,
                                     compress=self.compress_gdna)

    def download_gff3(self, logstream=sys.stderr):  # pragma: no cover
        """Download genome annotation."""
        subprocess.call(['mkdir', '-p', self.dbdir])
        if logstream is not None:
            logmsg = '[GenHub: %s] ' % self.config['species']
            logmsg += 'download genome annotation from %r' % self
            print(logmsg, file=logstream)
        genhub.download.url_download(self.gff3url, self.gff3path,
                                     compress=self.compress_gff3)

    def download_prot(self, logstream=sys.stderr):  # pragma: no cover
        """Download protein sequences."""
        subprocess.call(['mkdir', '-p', self.dbdir])
        if logstream is not None:
            logmsg = '[GenHub: %s] ' % self.config['species']
            logmsg += 'download protein sequences from %r' % self
            print(logmsg, file=logstream)
        genhub.download.url_download(self.proturl, self.protpath,
                                     compress=self.compress_prot)

    def download(self, logstream=sys.stderr):  # pragma: no cover
        """Run download task."""
        self.download_gdna(logstream)
        self.download_gff3(logstream)
        self.download_prot(logstream)

    def format(self, logstream=sys.stderr, verify=True):  # pragma: no cover
        """Run format task"""
        self.preprocess_gdna(logstream=logstream, verify=verify)
        self.preprocess_gff3(logstream=logstream, verify=verify)
        self.preprocess_prot(logstream=logstream, verify=verify)

    def preprocess(self, datatype, logstream=sys.stderr, verify=True):
        """
        Preprocess genome data files.

        To read from or write to a custom input/output streams (rather than the
        default file locations), set `instream` and/or `outstream` to a
        readable or writeable file handle, a stringio, or similar object.

        Note that this is a wrapper function: each subclass must implement 3
        methods (`format_gdna`, `format_gff3`, and `format_prot`) to do the
        actual formatting.
        """
        datatypes = {'gdna': 'genome sequence file',
                     'gff3': 'annotation file',
                     'prot': 'protein sequence file'}
        assert datatype in datatypes

        if logstream is not None:  # pragma: no cover
            logmsg = '[GenHub: %s] ' % self.config['species']
            logmsg += 'preprocess %s' % datatypes[datatype]
            print(logmsg, file=logstream)

        infile = {'gdna': self.gdnapath,
                  'gff3': self.gff3path,
                  'prot': self.protpath}[datatype]
        outfile = {'gdna': self.gdnafile,
                   'gff3': self.gff3file,
                   'prot': self.protfile}[datatype]
        if datatype != 'gff3':
            if infile.endswith('.gz'):
                instream = gzip.open(infile, 'rt')
            else:
                instream = open(infile, 'r')
            outstream = open(outfile, 'w')

        if datatype == 'gdna':
            self.format_gdna(instream, outstream, logstream)
        elif datatype == 'prot':
            self.format_prot(instream, outstream, logstream)
        else:
            self.format_gff3(logstream)

        if datatype != 'gff3':
            instream.close()
            outstream.close()

        if verify is False:
            return

        if 'checksums' in self.config and datatype in self.config['checksums']:
            sha1 = self.config['checksums'][datatype]
            testsha1 = self.file_sha1(outfile)
            assert testsha1 == sha1, ('%s %s integrity check failed\n%s\n%s' %
                                      (self.label, datatypes[datatype],
                                       testsha1, sha1))
        else:  # pragma: no cover
            message = 'Cannot verify integrity of %s ' % self.label
            message += '%s without a checksum' % datatypes[datatype]
            print(message, file=logstream)

    def preprocess_gdna(self, logstream=sys.stderr, verify=True):
        self.preprocess('gdna', logstream, verify)

    def preprocess_gff3(self, logstream=sys.stderr, verify=True):
        self.preprocess('gff3', logstream, verify)

    def preprocess_prot(self, logstream=sys.stderr, verify=True):
        self.preprocess('prot', logstream, verify)

    def filter_file(self):
        """
        Write exclusion filter to a temporary file and return.

        Data configurations may include an optional `annotfilter` with patterns
        to discard from the input annotation a la `grep -v`. This function
        retrieves the pattern(s) from the genome configuration and writes them
        to a temporary file that can be used in a `grep -vf` command. The
        calling function is responsible for unlinking the temporary file from
        the operating system.
        """
        if 'annotfilter' not in self.config:
            return None
        excludefile = tempfile.NamedTemporaryFile(mode='wt', delete=False)
        if isinstance(self.config['annotfilter'], str):
            print(self.config['annotfilter'], file=excludefile)
        else:
            for exclusion in self.config['annotfilter']:
                print(exclusion, file=excludefile)
        excludefile.close()
        return excludefile

    def file_sha1(self, filepath):
        """
        Stolen shamelessly from http://stackoverflow.com/a/19711609/459780.
        """
        sha = hashlib.sha1()
        with open(filepath, 'rb') as f:
            while True:
                block = f.read(2**10)
                if not block:
                    break
                sha.update(block)
            return sha.hexdigest()

    def cleanup(self, patterns_to_keep=None, fullclean=False, dryrun=False):
        """
        Clean up the DB working directory.

        By default, the files to be kept are the following.
        - *.iloci.fa
        - *.iloci.gff3
        - *.tsv
        - original (downloaded) data files
        All other files are deleted.

        If `fullclean` is true, the original data files are deleted as well.
        If `patterns_to_keep` is declared, each file to be deleted is checked
        to see if it contains any of the specified strings. If so, it is
        spared deletion.

        The `dryrun` parameter is just for unit testing.
        """
        dbfiles = glob.glob(self.dbdir + '/*')
        files_deleted = list()
        for dbfile in dbfiles:
            if dbfile.endswith('.iloci.fa') or dbfile.endswith('.iloci.gff3'):
                continue
            if dbfile.endswith('.tsv'):
                continue
            if dbfile.endswith('.protein2ilocus.txt'):
                continue
            if dbfile in [self.gdnapath, self.gff3path, self.protpath]:
                if not fullclean:
                    continue
            if patterns_to_keep:
                pattern_match = False
                for pattern in patterns_to_keep:
                    if pattern in dbfile:
                        pattern_match = True
                        break
                if pattern_match:
                    continue
            files_deleted.append(dbfile)
            if not dryrun:  # pragma: no cover
                os.unlink(dbfile)
        return files_deleted

    def get_prot_map(self):
        mapfile = '%s/%s.protein2ilocus.txt' % (self.dbdir, self.label)
        with open(mapfile, 'r') as instream:
            for line in instream:
                if line.strip() == '':  # pragma: no cover
                    continue
                protid, locid = line.strip().split()
                yield protid, locid


# -----------------------------------------------------------------------------
# Unit tests
# -----------------------------------------------------------------------------

def test_file_path():
    """GenomeDB File name resolution"""
    config = genhub.test_registry.genome('Bimp')
    db = GenomeDB('Bimp', config)
    assert db.file_path('bogus.txt') == './Bimp/bogus.txt'
    db = GenomeDB('Bimp', config, 'wd')
    assert db.file_path('Bimp.gff3') == 'wd/Bimp/Bimp.gff3'
    checkfailed = False
    try:
        config = genhub.test_registry.genome('Amel')
        db = GenomeDB('Amel', config)
        path = db.file_path('Amel.iloci.gff3', check=True)
    except FileNotFoundError as e:
        checkfailed = True
        assert e.args[0] == 'file "./Amel/Amel.iloci.gff3" not found'
    assert checkfailed


def test_props():
    """GenomeDB properties"""
    config = genhub.test_registry.genome('Bimp')
    db = GenomeDB('Bimp', config)
    assert db.dbdir == './Bimp'
    assert db.gdnafile == './Bimp/Bimp.gdna.fa'
    assert db.gff3file == './Bimp/Bimp.gff3'
    assert db.protfile == './Bimp/Bimp.all.prot.fa'
    assert db.source == 'refseq'

    config = genhub.test_registry.genome('Dqua')
    db = GenomeDB('Dqua', config, workdir='/opt/data/genomes')
    assert db.dbdir == '/opt/data/genomes/Dqua'
    assert db.gdnafile == '/opt/data/genomes/Dqua/Dqua.gdna.fa'
    assert db.gff3file == '/opt/data/genomes/Dqua/Dqua.gff3'
    assert db.protfile == '/opt/data/genomes/Dqua/Dqua.all.prot.fa'
    assert db.source == 'crg'


def test_filter_file():
    """GenomeDB filter file"""
    config = genhub.test_registry.genome('Lalb')
    db = GenomeDB('Lalb', config)
    assert db.filter_file() is None

    config = genhub.test_registry.genome('Drer')
    db = GenomeDB('Drer', config)
    ff = db.filter_file()
    with open(ff.name, 'r') as infile:
        excludestr = infile.read()
        assert excludestr.strip() == 'NC_002333.2'
    os.unlink(ff.name)
